# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import argparse, git, datetime, numpy, pygments.lexers, traceback, time, os, fnmatch, json, progressbar

def analyze():
    default_filetypes = set()
    for _, _, filetypes, _ in pygments.lexers.get_all_lexers():
        default_filetypes.update(filetypes)

    parser = argparse.ArgumentParser(description='Analyze git repo')
    parser.add_argument('--cohortfm', default='%Y', help='A Python datetime format string such as "%%Y" for creating cohorts (default: %(default)s)')
    parser.add_argument('--interval', default=7*24*60*60, type=int, help='Min difference between commits to analyze (default: %(default)s)')
    parser.add_argument('--ignore', default=[], action='append', help='File patterns that should be ignored (can provide multiple, will each subtract independently)')
    parser.add_argument('--only', default=[], action='append', help='File patterns that have to match (can provide multiple, will all have to match)')
    parser.add_argument('--outdir', default='.', help='Output directory to store results (default: %(default)s)')
    parser.add_argument('--branch', default='master', help='Branch to track (default: %(default)s)')
    parser.add_argument('--all-filetypes', action='store_true', help='Include all files (if not set then will only analyze %s' % ','.join(default_filetypes))
    parser.add_argument('repos', nargs=1)
    args = parser.parse_args()

    repo = git.Repo(args.repos[0])
    commit2cohort = {}
    code_commits = [] # only stores a subset
    master_commits = []
    commit2timestamp = {}
    curves_set = set()
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    print('Listing all commits')
    bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength)
    for i, commit in enumerate(repo.iter_commits(args.branch)):
        bar.update(i)
        cohort = datetime.datetime.utcfromtimestamp(commit.committed_date).strftime(args.cohortfm)
        commit2cohort[commit.hexsha] = cohort
        curves_set.add(('cohort', cohort))
        if len(commit.parents) == 1:
            code_commits.append(commit)
            last_date = commit.committed_date
            commit2timestamp[commit.hexsha] = commit.committed_date

    print('Backtracking the master branch')
    bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength)
    i, commit = 0, repo.head.commit
    last_date = None
    while True:
        bar.update(i)
        if not commit.parents:
            break
        if last_date is None or commit.committed_date < last_date - args.interval:
            master_commits.append(commit)
            last_date = commit.committed_date
        i, commit = i+1, commit.parents[0]

    ok_entry_paths = {}
    def entry_path_ok(path):
        # All this matching is slow so let's cache it
        if path not in ok_entry_paths:
            ok_entry_paths[path] = (
                (args.all_filetypes or any(fnmatch.fnmatch(os.path.split(path)[-1], filetype) for filetype in default_filetypes))
                and all([fnmatch.fnmatch(path, pattern) for pattern in args.only])
                and not any([fnmatch.fnmatch(path, pattern) for pattern in args.ignore]))
        return ok_entry_paths[path]

    def get_entries(commit):
        return [entry for entry in commit.tree.traverse()
                if entry.type == 'blob' and entry_path_ok(entry.path)]

    print('Counting total entries to analyze + caching filenames')
    entries_total = 0
    bar = progressbar.ProgressBar(max_value=len(master_commits))
    for i, commit in enumerate(reversed(master_commits)):
        bar.update(i)
        n = 0
        for entry in get_entries(commit):
            n += 1
            _, ext = os.path.splitext(entry.path)
            curves_set.add(('ext', ext))
            curves_set.add(('author', commit.author.name))
        entries_total += n

    def get_file_histogram(commit, path):
        h = {}
        try:
            for old_commit, lines in repo.blame(commit, path):
                cohort = commit2cohort.get(old_commit.hexsha, "MISSING")
                _, ext = os.path.splitext(path)
                keys = [('cohort', cohort), ('ext', ext), ('author', old_commit.author.name)]

                if old_commit.hexsha in commit2timestamp:
                    keys.append(('sha', old_commit.hexsha))

                for key in keys:
                    h[key] = h.get(key, 0) + len(lines)
        except KeyboardInterrupt:
            raise
        except:
            traceback.print_exc()
        return h

    curves = {}
    ts = []
    file_histograms = {}
    last_commit = None
    commit_history = {}

    print('Analyzing commit history')
    bar = progressbar.ProgressBar(max_value=entries_total, widget_kwargs=dict(samples=10000))
    entries_processed = 0
    for commit in reversed(master_commits):
        t = datetime.datetime.utcfromtimestamp(commit.committed_date)
        ts.append(t)
        changed_files = set()
        for diff in commit.diff(last_commit):
            if diff.a_blob:
                changed_files.add(diff.a_blob.path)
            if diff.b_blob:
                changed_files.add(diff.b_blob.path)
        last_commit = commit

        histogram = {}
        entries = get_entries(commit)
        for entry in entries:
            bar.update(entries_processed)
            entries_processed += 1
            if entry.path in changed_files or entry.path not in file_histograms:
                file_histograms[entry.path] = get_file_histogram(commit, entry.path)
            for key, count in file_histograms[entry.path].items():
                histogram[key] = histogram.get(key, 0) + count

        for key, count in histogram.items():
            key_type, key_item = key
            if key_type == 'sha':
                commit_history.setdefault(key_item, []).append((commit.committed_date, count))

        for key in curves_set:
            curves.setdefault(key, []).append(histogram.get(key, 0))


    def dump_json(output_fn, key_type, label_fmt=lambda x: x):
        key_items = sorted(k for t, k in curves_set if t == key_type)
        fn = os.path.join(args.outdir, output_fn)
        print('Writing %s data to %s' % (key_type, fn))
        f = open(fn, 'w')
        json.dump({'y': [curves[(key_type, key_item)] for key_item in key_items],
                   'ts': [t.isoformat() for t in ts],
                   'labels': [label_fmt(key_item) for key_item in key_items]
        }, f)
        f.close()

    # Dump accumulated stuff
    dump_json('cohorts.json', 'cohort', lambda c: 'Code added in %s' % c)
    dump_json('exts.json', 'ext')
    dump_json('authors.json', 'author')

    # Dump survival data
    fn = os.path.join(args.outdir, 'survival.json')
    f = open(fn, 'w')
    print('Writing survival data to %s' % fn)
    json.dump(commit_history, f)
    f.close()


if __name__ == '__main__':
    analyze()



# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib
matplotlib.use('Agg')

import sys, seaborn, dateutil.parser, numpy, json, collections, math, scipy.optimize, argparse, os

from matplotlib import pyplot

def survival_plot():
    parser = argparse.ArgumentParser(description='Plot survival plot')
    parser.add_argument('--exp-fit', action='store_true', help='Plot exponential fit')
    parser.add_argument('--display', action='store_true', help='Display plot')
    parser.add_argument('--outfile', default='survival_plot.png', help='Output file to store results (default: %(default)s)')
    parser.add_argument('--years', type=float, default=5, help='Number of years on x axis (default: %(default)s)')
    parser.add_argument('inputs', nargs='*')
    args = parser.parse_args()

    all_deltas = []
    YEAR = 365.25 * 24 * 60 * 60
    pyplot.figure(figsize=(13, 8))

    for fn in args.inputs:
        print('reading %s' % fn)
        commit_history = json.load(open(fn))

        print('counting %d commits' % len(commit_history))
        deltas = collections.defaultdict(lambda: numpy.zeros(2))
        total_n = 0
        for commit, history in commit_history.items():
            t0, orig_count = history[0]
            total_n += orig_count
            last_count = orig_count
            for t, count in history[1:]:
                deltas[t-t0] += (count-last_count, 0)
                last_count = count
            deltas[history[-1][0] - t0] += (-last_count, -orig_count)

        all_deltas.append((total_n, deltas))
        print('adding %d deltas...' % len(deltas))
        total_k = total_n
        P = 1.0
        xs = []
        ys = []
        for t in sorted(deltas.keys()):
            delta_k, delta_n = deltas[t]
            xs.append(t / YEAR)
            ys.append(100. * P)
            P *= 1 + delta_k / total_n
            total_k += delta_k
            total_n += delta_n
            if P < 0.05:
                break

        print('plotting...')
        if args.exp_fit:
            pyplot.plot(xs, ys, color='darkgray')
        else:
            parts = os.path.split(fn)
            pyplot.plot(xs, ys, label=(len(parts) > 1 and parts[-2] or None))


    def fit(k):
        loss = 0.0
        for total_n, deltas in all_deltas:
            total_k = total_n
            P = 1.0
            for t in sorted(deltas.keys()):
                delta_k, delta_n = deltas[t]
                pred = total_n * math.exp(-k * t / YEAR)
                loss += (total_n * P - pred)**2
                P *= 1 + delta_k / total_n
                total_k += delta_k
                total_n += delta_n
        print(k, loss)
        return loss

    if args.exp_fit:
        print('fitting exponential function')
        k = scipy.optimize.fmin(fit, 0.5, maxiter=50)[0]
        ts = numpy.linspace(0, args.years, 1000)
        ys = [100. * math.exp(-k * t) for t in ts]
        pyplot.plot(ts, ys, color='red', label='Exponential fit, half-life = %.2f years' % (math.log(2) / k))

    pyplot.xlabel('Years')
    pyplot.ylabel('%')
    pyplot.xlim([0, args.years])
    pyplot.ylim([0, 100])
    pyplot.title('% of lines still present in code after n years')
    pyplot.legend()
    pyplot.tight_layout()
    pyplot.savefig(args.outfile)
    if args.display:
        pyplot.show()


if __name__ == '__main__':
    survival_plot()



# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib
matplotlib.use('Agg')

import argparse, dateutil.parser, itertools, json, numpy, seaborn, sys
from matplotlib import pyplot


def generate_n_colors(n):
    vs = numpy.linspace(0.4, 1.0, 7)
    colors = [(.9, .4, .4)]
    def euclidean(a, b):
        return sum((x-y)**2 for x, y in zip(a, b))
    while len(colors) < n:
        new_color = max(itertools.product(vs, vs, vs), key=lambda a: min(euclidean(a, b) for b in colors))
        colors.append(new_color)
    return colors


def stack_plot():
    parser = argparse.ArgumentParser(description='Plot stack plot')
    parser.add_argument('--display', action='store_true', help='Display plot')
    parser.add_argument('--outfile', default='stack_plot.png', help='Output file to store results (default: %(default)s)')
    parser.add_argument('--max-n', default=20, type=int, help='Max number of dataseries (will roll everything else into "other") (default: %(default)s)')
    parser.add_argument('--normalize', action='store_true', help='Normalize the plot to 100%')
    parser.add_argument('inputs')
    args = parser.parse_args()

    data = json.load(open(args.inputs))
    y = numpy.array(data['y'])
    if y.shape[0] > args.max_n:
        js = sorted(range(len(data['labels'])), key=lambda j: max(y[j]), reverse=True)
        other_sum = numpy.sum(y[j] for j in js[args.max_n:])
        top_js = sorted(js[:args.max_n], key=lambda j: data['labels'][j])
        y = numpy.array([y[j] for j in top_js] + [other_sum])
        labels = [data['labels'][j] for j in top_js] + ['other']
    else:
        labels = data['labels']
    if args.normalize:
        y = 100. * numpy.array(y) / numpy.sum(y, axis=0)
    pyplot.figure(figsize=(13, 8))
    pyplot.stackplot([dateutil.parser.parse(t) for t in data['ts']],
                     numpy.array(y),
                     labels=labels,
                     colors=generate_n_colors(len(labels)))
    pyplot.legend(loc=2)
    if args.normalize:
        pyplot.ylabel('Share of lines of code (%)')
        pyplot.ylim([0, 100])
    else:
        pyplot.ylabel('Lines of code')
    pyplot.savefig(args.outfile)
    pyplot.tight_layout()
    if args.display:
        pyplot.show()


if __name__ == '__main__':
    stack_plot()



from distutils.core import setup

setup(name='git-of-theseus',
      version='0.1.2',
      description='Plot stats on Git repositories',
      author='Erik Bernhardsson',
      author_email='mail@erikbern.com',
      url='https://github.com/erikbern/git-of-theseus',
      packages=['git_of_theseus'],
      scripts=['scripts/git-of-theseus-analyze', 'scripts/git-of-theseus-stack-plot', 'scripts/git-of-theseus-survival-plot'],
      install_requires=[
          'gitpython',
          'numpy',
          'progressbar2',
          'pygments',
          'matplotlib',
          'seaborn',
      ]
)
